import http.client
import socket
from base64 import b64encode
import json
from . import resources
from .resource import Resource, Empty
from .request import Request
from .response import Response
from recurly import USER_AGENT, ApiError, NetworkError
from pydoc import locate
import urllib.parse
from datetime import datetime

PORT = 443
HOST = "v3.recurly.com"
BINARY_TYPES = ["application/pdf"]


def request_converter(value):
    """Used by json serializer to cast values"""
    if isinstance(value, datetime):
        return value.isoformat()
    else:
        return value


class BaseClient:
    def __init__(self, api_key):
        self.__api_key = api_key
        self.__conn = http.client.HTTPSConnection(HOST, PORT)

    def _make_request(self, method, path, body, params):
        try:
            basic_auth = b64encode(bytes(self.__api_key + ":", "ascii")).decode("ascii")
            headers = {
                "User-Agent": USER_AGENT,
                "Authorization": "Basic %s" % basic_auth,
                "Accept": "application/vnd.recurly.%s" % self.api_version(),
                "Content-Type": "application/json",
            }
            if body:
                body = json.dumps(body, default=request_converter)

            if params:
                path += "?" + self._url_encode(params)

            self.__conn.request(method, path, body, headers=headers)
            request = Request(method, path, body)
            resp = Response(self.__conn.getresponse(), request)

            if resp.status >= 400:
                if resp.body:
                    raise Resource.cast_error(resp)
                else:
                    raise ApiError(
                        "Unknown Error. Recurly Request Id: " + str(resp.request_id),
                        None,
                    )

            if resp.body:
                if resp.content_type in BINARY_TYPES:
                    return Resource.cast_file(resp)
                else:
                    json_body = json.loads(resp.body.decode("utf-8"))
                    return Resource.cast_json(json_body, response=resp)
            else:
                return Resource.cast_json({}, Empty, resp)

        except socket.error as e:
            raise NetworkError(e)

    def _validate_path_parameters(self, args):
        """Checks that path parameters are valid"""
        # Check that parameters are valid types
        if any(type(arg) not in [str, int, float] for arg in args):
            raise ApiError("Invalid parameter type", None)

        # Check that string parameters are not empty
        if any(isinstance(arg, str) and not bool(arg.strip()) for arg in args):
            raise ApiError("Parameters cannot be empty strings", None)

    def _interpolate_path(self, path, *args):
        """Encodes components and interpolates path"""
        self._validate_path_parameters(args)

        return path % tuple(map(lambda arg: urllib.parse.quote(arg, safe=""), args))

    def _url_encode(self, params):
        """Encode query params for URL. We need to customize this to conform to Recurly's API"""
        r_params = {}

        for k, v in params.items():
            # join lists w/ a comma (CSV encoding)
            if isinstance(v, list) or isinstance(v, tuple):
                r_params[k] = ",".join(v)
            # booleans need to be downcased
            elif isinstance(v, bool):
                r_params[k] = "true" if v else "false"
            # datetimes should be iso8601 strings
            elif isinstance(v, datetime):
                r_params[k] = v.isoformat()
            else:
                r_params[k] = v

        return urllib.parse.urlencode(r_params)



from datetime import datetime
import recurly


def parse_int_header(dictionary, key):
    """Defensively parse an int header"""
    val = dictionary.get(key)

    if val is None or not isinstance(val, str):
        return None

    if not val.isdigit():
        return None

    return int(val)


def parse_datetime_header(dictionary, key):
    """Defensively parse datetime header"""

    int_val = parse_int_header(dictionary, key)
    if int_val is None:
        return None

    return datetime.utcfromtimestamp(int_val)


class Response:
    """Class representing a response from Recurly"""

    def __init__(self, response, request):
        self.request = request
        self.status = response.status
        http_body = response.read()
        self.body = None
        self.__headers = response.headers

        self.request_id = self.__headers.get("X-Request-Id")
        self.date = self.__headers.get("Date")

        self.rate_limit = parse_int_header(self.__headers, "X-RateLimit-Limit")
        self.rate_limit_remaining = parse_int_header(
            self.__headers, "X-RateLimit-Remaining"
        )
        self.rate_limit_reset = parse_datetime_header(
            self.__headers, "X-RateLimit-Reset"
        )
        self.content_type = self.__headers.get("Content-Type", "").split(";")[0]

        self.proxy_metadata = {
            "server": self.__headers.get("Server"),
            "cf-ray": self.__headers.get("CF-RAY"),
        }

        self.total_records = parse_int_header(self.__headers, "Recurly-Total-Records")

        if http_body and len(http_body) > 0:
            self.body = http_body



import unittest
import recurly
from datetime import datetime
from recurly import Response, Request
from unittest.mock import Mock, MagicMock


class TestResponse(unittest.TestCase):
    def test_init(self):
        resp = MagicMock()
        resp.headers = {
            "X-Request-Id": "0av50sm5l2n2gkf88ehg",
            "X-RateLimit-Limit": "2000",
            "X-RateLimit-Remaining": "1985",
            "X-RateLimit-Reset": "1564624560",
            "Recurly-Total-Records": "100",
            "Date": "Thu, 01 Aug 2019 01:26:44 GMT",
            "Server": "cloudflare",
            "CF-RAY": "4ff4b71268424738-EWR",
        }

        req = Request("GET", "/sites", {})
        response = Response(resp, req)

        self.assertEqual(type(response), Response)
        self.assertEqual(response.request_id, "0av50sm5l2n2gkf88ehg")
        self.assertEqual(response.rate_limit, 2000)
        self.assertEqual(response.rate_limit_remaining, 1985)
        self.assertEqual(response.rate_limit_reset, datetime(2019, 8, 1, 1, 56))
        self.assertEqual(response.total_records, 100)
        self.assertEqual(response.date, "Thu, 01 Aug 2019 01:26:44 GMT")
        self.assertEqual(response.proxy_metadata["server"], "cloudflare")
        self.assertEqual(response.proxy_metadata["cf-ray"], "4ff4b71268424738-EWR")
        self.assertEqual(response.request.method, "GET")
        self.assertEqual(response.request.path, "/sites")
        self.assertEqual(response.request.body, {})

    def test_init_with_invalid_headers(self):
        resp = MagicMock()
        resp.headers = {
            "X-Request-Id": "0av50sm5l2n2gkf88ehg",
            "X-RateLimit-Limit": "notanum",
            "X-RateLimit-Remaining": "notanum",
            "X-RateLimit-Reset": "notanum",
            "recurly-total-records": "notanum",
            "Date": "Thu, 01 Aug 2019 01:26:44 GMT",
            "Server": "cloudflare",
            "CF-RAY": "4ff4b71268424738-EWR",
        }

        req = Request("GET", "/sites", {})
        response = Response(resp, req)

        self.assertEqual(type(response), Response)
        self.assertEqual(response.request_id, "0av50sm5l2n2gkf88ehg")
        self.assertEqual(response.rate_limit, None)
        self.assertEqual(response.rate_limit_remaining, None)
        self.assertEqual(response.rate_limit_reset, None)
        self.assertEqual(response.total_records, None)
        self.assertEqual(response.date, "Thu, 01 Aug 2019 01:26:44 GMT")
        self.assertEqual(response.proxy_metadata["server"], "cloudflare")
        self.assertEqual(response.proxy_metadata["cf-ray"], "4ff4b71268424738-EWR")

    def test_init_with_missing_headers(self):
        resp = MagicMock()
        resp.headers = {}

        req = Request("GET", "/sites", {})
        response = Response(resp, req)

        self.assertEqual(type(response), Response)
        self.assertEqual(response.request_id, None)
        self.assertEqual(response.rate_limit, None)
        self.assertEqual(response.rate_limit_remaining, None)
        self.assertEqual(response.rate_limit_reset, None)
        self.assertEqual(response.total_records, None)
        self.assertEqual(response.date, None)
        self.assertEqual(response.proxy_metadata["server"], None)
        self.assertEqual(response.proxy_metadata["cf-ray"], None)



import unittest
import recurly
from inspect import getmembers, isclass
import datetime


resources = dict(
    [
        m
        for m in getmembers(recurly.resources)
        if isclass(m[1])
        and issubclass(m[1], recurly.Resource)
        and m[1] is not recurly.Resource
    ]
)


class TestResources(unittest.TestCase):
    def test_schemas(self):
        n_resources = len(resources)
        self.assertTrue(n_resources > 1)

    def test_has_error(self):
        error_class = resources["Error"]
        self.assertTrue(issubclass(error_class, recurly.Resource))

    def test_schemas(self):
        allowed_primitives = [str, bool, int, float, dict, list, datetime]
        for name, klass in resources.items():
            schema = klass.schema
            self.assertIsInstance(schema, dict)
            for k, v in schema.items():
                if isinstance(v, str):
                    subclass = resources[v]
                    self.assertTrue(
                        issubclass(subclass, recurly.Resource),
                        "You can only associate with other Resources",
                    )
                elif isinstance(v, list) and len(v) == 1 and isinstance(v[0], str):
                    subclass = resources[v[0]]
                    self.assertTrue(
                        issubclass(subclass, recurly.Resource),
                        "You can only associate with other Resources",
                    )
                else:
                    self.assertIn(
                        v,
                        allowed_primitives,
                        "%s has invalid schema definition for %s" % (name, k),
                    )



from recurly.base_client import BaseClient
from recurly.pager import Pager


class MockClient(BaseClient):
    def api_version(self):
        return "v2018-08-09"

    def list_resources(self, params={}):
        path = "/resources"
        return Pager(self, path, params)

    def get_resource(self, resource_id, params={}):
        path = self._interpolate_path("/resources/%s", resource_id)
        return self._make_request("GET", path, None, params)

    def update_resource(self, resource_id, body):
        path = self._interpolate_path("/resources/%s", resource_id)
        return self._make_request("PUT", path, body, None)

    def delete_resource(self, resource_id):
        path = self._interpolate_path("/resources/%s", resource_id)
        return self._make_request("DELETE", path, None, None)



import setuptools
import re
import os.path

# get version from top level __init__ file
VERSION_RE = re.compile('^__version__ = "(.+)"$', flags=re.MULTILINE)
with open(os.path.join(os.path.dirname(__file__), "recurly", "__init__.py")) as PACKAGE:
    VERSION = VERSION_RE.search(PACKAGE.read()).group(1)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="recurly",
    # version managed in recurly/__init__.py
    version=VERSION,
    author="Benjamin Eckel",
    author_email="ben@recurly.com",
    description="Recurly v3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/recurly/recurly-client-python",
    packages=setuptools.find_packages(),
    classifiers=[],
    tests_require=["coverage"],
)



#
# This file is automatically created by Recurly's OpenAPI generation process
# and thus any edits you make by hand will be lost. If you wish to make a
# change to this file, please create a Github issue explaining the changes you
# need and we will usher them to the appropriate places.
import recurly

ERROR_MAP = {
    500: "InternalServerError",
    502: "BadGatewayError",
    503: "ServiceUnavailableError",
    304: "NotModifiedError",
    400: "BadRequestError",
    401: "UnauthorizedError",
    402: "PaymentRequiredError",
    403: "ForbiddenError",
    404: "NotFoundError",
    406: "NotAcceptableError",
    412: "PreconditionFailedError",
    422: "UnprocessableEntityError",
    429: "TooManyRequestsError",
}


class ResponseError(recurly.ApiError):
    pass


class ServerError(ResponseError):
    pass


class InternalServerError(ServerError):
    pass


class BadGatewayError(ServerError):
    pass


class ServiceUnavailableError(ServerError):
    pass


class RedirectionError(ResponseError):
    pass


class NotModifiedError(ResponseError):
    pass


class ClientError(recurly.ApiError):
    pass


class BadRequestError(ClientError):
    pass


class InvalidContentTypeError(BadRequestError):
    pass


class UnauthorizedError(ClientError):
    pass


class PaymentRequiredError(ClientError):
    pass


class ForbiddenError(ClientError):
    pass


class InvalidApiKeyError(ForbiddenError):
    pass


class InvalidPermissionsError(ForbiddenError):
    pass


class NotFoundError(ClientError):
    pass


class NotAcceptableError(ClientError):
    pass


class UnknownApiVersionError(NotAcceptableError):
    pass


class UnavailableInApiVersionError(NotAcceptableError):
    pass


class InvalidApiVersionError(NotAcceptableError):
    pass


class PreconditionFailedError(ClientError):
    pass


class UnprocessableEntityError(ClientError):
    pass


class ValidationError(UnprocessableEntityError):
    pass


class MissingFeatureError(UnprocessableEntityError):
    pass


class TransactionError(UnprocessableEntityError):
    pass


class SimultaneousRequestError(UnprocessableEntityError):
    pass


class ImmutableSubscriptionError(UnprocessableEntityError):
    pass


class InvalidTokenError(UnprocessableEntityError):
    pass


class TooManyRequestsError(ClientError):
    pass


class RateLimitedError(TooManyRequestsError):
    pass
